(* parser/parse.sml *)

functor ParseFn(structure Absyn : ABSYN
		structure RMLParser : ARG_PARSER
				      where type arg = unit
					and type pos = int
		structure Tokens : RML_TOKENS
		structure LexArg : LEXARG
		sharing type Tokens.token = RMLParser.Token.token
		sharing type Tokens.svalue = RMLParser.svalue
		sharing type RMLParser.result = Absyn.module
		sharing type RMLParser.lexarg = LexArg.lexarg
		sharing Absyn.Source = LexArg.Source
		  ) : PARSE =
  struct
    structure Absyn = Absyn

    fun parse(startToken, file) =
      let val is = TextIO.openIn file
      in
	(let val (lexarg,inputf) = LexArg.new(file, is)
	     val pos = 2	(*XXX: ML-Lex*)
	     val lexer = RMLParser.makeLexer inputf lexarg
	     val lexer = RMLParser.Stream.cons(startToken(pos,pos), lexer)
	     val (result,_) = RMLParser.parse(0,lexer,LexArg.error2 lexarg,())
	 in
	   if LexArg.seenErr lexarg then raise RMLParser.ParseError else ();
	   TextIO.closeIn is;
	   let val Absyn.MODULE(Absyn.INTERFACE{modid,specs,...},decs) = result
	       val interface = Absyn.INTERFACE{modid=modid,specs=specs,
					       source=LexArg.source lexarg}
	   in
	     Absyn.MODULE(interface, decs)
	   end
	 end) handle e => (TextIO.closeIn is; raise e)
      end

    fun parseModule file = parse(Tokens.START_MODULE, file)
    fun parseInterface file = parse(Tokens.START_INTERFACE, file)

  end (* functor ParseFn *)
